local args = {};

--- @alias consumer (fun(next: fun(): string?)) | string
--- @alias consumers { [string]: consumer, [1]: consumer }

--- @param consumers consumers
--- @param name string
local function get_consumer(consumers, name)
	local consumer = nil;
	local path = {};
	local n = 0;

	while true do
		local curr = consumer or name;
		if path[curr] then
			local path_arr = {};

			for k, v in next, path do
				path_arr[v] = k;
			end

			error("Alias to '" .. curr .. "' is recursive: " .. table.concat(path_arr, " -> "));
		end

		consumer = consumers[curr];
		if consumer == nil then
			local prefix;
			if n == 0 then
				prefix = "Unknown flag";
			else
				prefix = "Unknown alias";
			end

			if #curr == 1 then
				error(prefix .. " '-" .. curr .. "'");
			else
				error(prefix .. " '--" .. curr .. "'");
			end
		elseif type(consumer) == "function" then
			return consumer;
		end

		path[curr] = n;
		n = n + 1;
	end
end

local digest_next;

--- @param consumers consumers
--- @param name string
--- @param next fun(): string?
local function digest_arg(consumers, name, next)
	get_consumer(consumers, name)(next);
	return digest_next(consumers, next);
end

--- @param consumers consumers
--- @param next fun(): string?
--- @param only_rest? boolean
local function digest_rest(consumers, next, only_rest)
	if not consumers[1] then
		local arg = next();
		if not arg then return nil end
		error("Invalid argument '" .. arg .. "'");
	end

	local consumed = false;
	local function our_consumer()
		consumed = true;
		return next();
	end

	consumers[1](our_consumer);

	if not consumed then
		local arg = next();
		if not arg then return nil end
		error("Invalid argument '" .. arg .. "'");
	end

	if only_rest then
		return digest_rest(consumers, next, true);
	else
		return digest_next(consumers, next);
	end
end

--- @param consumers consumers
--- @param next fun(): string?
function digest_next(consumers, next)
	local arg = next();
	if not arg then return nil end

	if arg == "--" then
		return digest_rest(consumers, next, true);
	end

	local name, val;

	name, val = arg:match "^%-%-(.-)=(.+)$";
	if name then
		local function our_next()
			our_next = next;
			return val;
		end

		return digest_arg(consumers, name, function () return our_next() end);
	end

	name = arg:match "^%-%-(.-)$";
	if name then
		return digest_arg(consumers, name, next);
	end

	name = arg:match "^%-(.-)$";
	if name then
		for c in name:gmatch "." do
			digest_arg(consumers, c, next);
		end

		return false;
	end

	local function our_next()
		our_next = next;
		return arg;
	end

	return digest_rest(consumers, function () return our_next() end, false);
end

--- @param consumers consumers
--- @returns fun(...: string)
function args.parser(consumers)
	return function (...)
		local args = { ... };
		local i = 0;

		local function next()
			i = i + 1;
			return args[i];
		end

		return digest_next(consumers, next);
	end
end

return args;
