local utils = require "utils";
local c = {};

local function c_escape(str)
	return "\"" .. str:gsub(".", function (c)
		local b = string.byte(c);

		if c == "\n" then
			return "\\n";
		elseif c == "\\" then
			return "\\\\";
		elseif c == "\"" then
			return "\\\"";
		elseif b >= 32 and b <= 126 then
			return c;
		else
			return ("\\%.3o"):format(b);
		end
	end) .. "\"";
end

--- Executes a GCC-like C compiler
--- @param settings { inputs: string[], output: string, flags?: string[], cc?: string }
--- @return task
function c.compile(settings)
	local cc = settings.cc or "cc";
	local inputs = settings.inputs;
	local output = settings.output;
	local flags = settings.flags or {};

	return function ()
		if utils.check_files(inputs, { output }) then return end

		local rest = {};
		table.move(inputs, 1, #inputs, 1, rest);
		table.move(flags, 1, #flags, #rest + 1, rest);

		return utils.simple_exec(cc, "-o", output, table.unpack(rest));
	end
end

--- Creates a C file, defining the specified constant that will contain a string representation of the input file
--- @param settings { input: string, output: string, const_name: string }
--- @return task
function c.asset(settings)
	local input = settings.input;
	local output = settings.output;
	local const_name = settings.const_name;

	return function ()
		if utils.check_files({ input }, { output }) then return end

		local out_f = assert(io.open(output, "w"));

		assert(out_f:write "#pragma once\n");
		assert(out_f:write "unsigned char \n");
		assert(out_f:write(const_name));
		assert(out_f:write " = ");

		local has_any = false;

		for line in assert(io.lines(input, 128)) do
			has_any = true;
			assert(out_f:write "\n\t");
			assert(out_f:write(c_escape(line)));
		end

		if not has_any then
			assert(out_f:write "\"\"");
		end

		assert(out_f:write ";\n");
		assert(out_f:close());
	end
end

return c;
