--- @class recurse_state
--- @field stack ({ prefix: string, data: any, next: fun(): string? } | string)[]
--- @field readdir fun(dir: string): (fun(state: any): string?)

--- @param state recurse_state
--- @return string?
local function recurser(state)
	if not state then
		return nil;
	end

	while true do
		local top = table.remove(state.stack);
		if not top then
			return nil;
		elseif type(top) == "string" then
			local res = state.readdir(top);

			if res then
				table.insert(state.stack, { next = res, prefix = top .. "/" });
			end

			return top;
		elseif type(top) == "table" then
			local res = top.next(top.data);
			if res then
				table.insert(state.stack, top.prefix .. res);
				table.insert(state.stack, top);
			end
		else
			error "invalid stack state";
		end
	end
end

return function (dir, readdir)
	return recurser, { stack = { dir }, readdir = readdir or require "fs".readdir };
end
